@@ -9,24 +9,21 @@ Cu.import("resource://gre/modules/XPCOMUtils.jsm");
 Cu.import("resource://gre/modules/Services.jsm");
 Cu.import("resource://gre/modules/NetUtil.jsm");
 
-const DEBUG = false;
+const DEBUG = Services.prefs.getBoolPref("dom.presentation.tcp_server.debug");
 function log(aMsg) {
   dump("-*- TCPPresentationServer.js: " + aMsg + "\n");
 }
 
-function TCPDeviceInfo(aHost, aPort, aId, aName, aType) {
-  this.host = aHost;
+function TCPDeviceInfo(aAddress, aPort, aId) {
+  this.address = aAddress;
   this.port = aPort;
   this.id = aId;
-  this.name = aName;
-  this.type = aType;
 }
 
 function TCPPresentationServer() {
   this._id = null;
   this._port = 0;
   this._serverSocket = null;
-  this._devices = new Map(); // id -> device
 }
 
 TCPPresentationServer.prototype = {
@@ -37,41 +34,43 @@ TCPPresentationServer.prototype = {
    */
   _controlChannels: [],
 
-  init: function(aId, aPort) {
-    if (this._isInit()) {
+  startService: function(aPort) {
+    if (this._isServiceInit()) {
       DEBUG && log("TCPPresentationServer - server socket has been initialized");
       throw Cr.NS_ERROR_FAILURE;
     }
 
-    if (typeof aId === "undefined" || typeof aPort === "undefined") {
-      DEBUG && log("TCPPresentationServer - aId/aPort should not be undefined");
-      throw Cr.NS_ERROR_FAILURE;
-    }
-
-    DEBUG && log("TCPPresentationServer - init id: " + aId + " port: " + aPort);
-
     /**
      * 0 or undefined indicates opt-out parameter, and a port will be selected
      * automatically.
      */
-    let serverSocketPort = (aPort !== 0) ? aPort : -1;
+    let serverSocketPort = (typeof aPort !== "undefined" && aPort !== 0) ? aPort : -1;
 
     this._serverSocket = Cc["@mozilla.org/network/server-socket;1"]
                          .createInstance(Ci.nsIServerSocket);
+
+    if (!this._serverSocket) {
+      DEBUG && log("TCPPresentationServer - create server socket fail.");
+      throw Cr.NS_ERROR_FAILURE;
+    }
+
     try {
       this._serverSocket.init(serverSocketPort, false, -1);
+      this._serverSocket.asyncListen(this);
     } catch (e) {
       // NS_ERROR_SOCKET_ADDRESS_IN_USE
       DEBUG && log("TCPPresentationServer - init server socket fail: " + e);
       throw Cr.NS_ERROR_FAILURE;
     }
 
-    /**
-     * The setter may trigger |_serverSocket.asyncListen| if the |id| setting
-     * successes.
-     */
-    this.id = aId;
     this._port = this._serverSocket.port;
+
+    DEBUG && log("TCPPresentationServer - service start on port: " + this._port);
+
+    // Monitor network interface change to restart server socket.
+    // Only B2G has nsINetworkManager
+    Services.obs.addObserver(this, "network-active-changed", false);
+    Services.obs.addObserver(this, "network:offline-status-changed", false);
   },
 
   get id() {
@@ -79,16 +78,7 @@ TCPPresentationServer.prototype = {
   },
 
   set id(aId) {
-    if (!aId || aId.length == 0 || aId === this._id) {
-      return;
-    } else if (this._id) {
-      throw Cr.NS_ERROR_FAILURE;
-    }
     this._id = aId;
-
-    if (this._serverSocket) {
-      this._serverSocket.asyncListen(this);
-    }
   },
 
   get port() {
@@ -103,52 +93,16 @@ TCPPresentationServer.prototype = {
     return this._listener;
   },
 
-  _isInit: function() {
-    return this._id !== null && this._serverSocket !== null;
+  _isServiceInit: function() {
+    return this._serverSocket !== null;
   },
 
-  close: function() {
-    DEBUG && log("TCPPresentationServer - close");
-    if (this._serverSocket) {
-      this._serverSocket.close();
-    }
-    this._id = null;
-    this._port = 0;
-  },
-
-  createTCPDevice: function(aId, aName, aType, aHost, aPort) {
-    if (this._devices.has(aId)) {
-      throw Cr.NS_ERROR_INVALID_ARG;
-    }
-
-    this._devices.set(aId, new TCPDevice(this, {id: aId,
-                                                name: aName,
-                                                type: aType,
-                                                host: aHost,
-                                                port: aPort}));
-    return this._devices.get(aId);
-  },
-
-  getTCPDevice: function(aId) {
-    if (!this._devices.has(aId)) {
-      throw Cr.NS_ERROR_INVALID_ARG;
-    }
-    return this._devices.get(aId);
-  },
-
-  removeTCPDevice: function(aId) {
-    if (!this._devices.has(aId)) {
-      throw Cr.NS_ERROR_INVALID_ARG;
-    }
-    this._devices.delete(aId);
-  },
-
-  requestSession: function(aDevice, aUrl, aPresentationId) {
-    if (!this._isInit()) {
-      DEBUG && log("TCPPresentationServer - has not initialized; requestSession fails");
+  requestSession: function(aDeviceInfo, aUrl, aPresentationId) {
+    if (!this.id) {
+      DEBUG && log("TCPPresentationServer - Id has not initialized; requestSession fails");
       return null;
     }
-    DEBUG && log("TCPPresentationServer - requestSession to " + aDevice.name
+    DEBUG && log("TCPPresentationServer - requestSession to " + aDeviceInfo.id
                  + ": " + aUrl + ", " + aPresentationId);
 
     let sts = Cc["@mozilla.org/network/socket-transport-service;1"]
@@ -158,8 +112,8 @@ TCPPresentationServer.prototype = {
     try {
       socketTransport = sts.createTransport(null,
                                             0,
-                                            aDevice.host,
-                                            aDevice.port,
+                                            aDeviceInfo.address,
+                                            aDeviceInfo.port,
                                             null);
     } catch (e) {
       DEBUG && log("TCPPresentationServer - createTransport throws: " + e);
@@ -168,22 +122,23 @@ TCPPresentationServer.prototype = {
     }
     return new TCPControlChannel(this,
                                  socketTransport,
-                                 aDevice,
+                                 aDeviceInfo,
                                  aPresentationId,
                                  "sender",
                                  aUrl);
   },
 
-  responseSession: function(aDevice, aSocketTransport) {
-    if (!this._isInit()) {
-      DEBUG && log("TCPPresentationServer - has not initialized; responseSession fails");
+  responseSession: function(aDeviceInfo, aSocketTransport) {
+    if (!this._isServiceInit()) {
+      DEBUG && log("TCPPresentationServer - should never receive remote " +
+                   "session request before server socket initialization");
       return null;
     }
     DEBUG && log("TCPPresentationServer - responseSession to "
-                 + JSON.stringify(aDevice));
+                 + JSON.stringify(aDeviceInfo));
     return new TCPControlChannel(this,
                                  aSocketTransport,
-                                 aDevice,
+                                 aDeviceInfo,
                                  null, // presentation ID
                                  "receiver",
                                  null // url
@@ -191,18 +146,13 @@ TCPPresentationServer.prototype = {
   },
 
   // Triggered by TCPControlChannel
-  onSessionRequest: function(aId, aUrl, aPresentationId, aControlChannel) {
-    let device = this._devices.get(aId);
-    if (!device) {
-      //XXX Bug 1136565 - should have a way to recovery
-      DEBUG && log("TCPPresentationServer - onSessionRequest not found device for id: "
-                   + aId );
-      return;
-    }
-    device.listener.onSessionRequest(device,
-                                     aUrl,
-                                     aPresentationId,
-                                     aControlChannel);
+  onSessionRequest: function(aDeviceInfo, aUrl, aPresentationId, aControlChannel) {
+    DEBUG && log("TCPPresentationServer - onSessionRequest: "
+                 + aDeviceInfo.address + ":" + aDeviceInfo.port);
+    this.listener.onSessionRequest(aDeviceInfo,
+                                   aUrl,
+                                   aPresentationId,
+                                   aControlChannel);
     this.releaseControlChannel(aControlChannel);
   },
 
@@ -210,8 +160,8 @@ TCPPresentationServer.prototype = {
   onSocketAccepted: function(aServerSocket, aClientSocket) {
     DEBUG && log("TCPPresentationServer - onSocketAccepted: "
                  + aClientSocket.host + ":" + aClientSocket.port);
-    let device = new TCPDeviceInfo(aClientSocket.host, aClientSocket.port);
-    this.holdControlChannel(this.responseSession(device, aClientSocket));
+    let deviceInfo = new TCPDeviceInfo(aClientSocket.host, aClientSocket.port);
+    this.holdControlChannel(this.responseSession(deviceInfo, aClientSocket));
   },
 
   holdControlChannel: function(aControlChannel) {
@@ -228,30 +178,74 @@ TCPPresentationServer.prototype = {
   // nsIServerSocketListener (Triggered by nsIServerSocket.init)
   onStopListening: function(aServerSocket, aStatus) {
     DEBUG && log("TCPPresentationServer - onStopListening: " + aStatus);
-
-    // manually closed
-    if (aStatus === Cr.NS_BINDING_ABORTED) {
-      aStatus = Cr.NS_OK;
-    }
-    this._listener && this._listener.onClose(aStatus);
-    this._serverSocket = null;
   },
 
   close: function() {
-    DEBUG && log("TCPPresentationServer - close signalling channel");
-    if (this._serverSocket) {
+    DEBUG && log("TCPPresentationServer - close");
+    if (this._isServiceInit()) {
+      DEBUG && log("TCPPresentationServer - close server socket");
       this._serverSocket.close();
       this._serverSocket = null;
+
+      Services.obs.removeObserver(this, "network-active-changed");
+      Services.obs.removeObserver(this, "network:offline-status-changed");
     }
-    this._id = null;
     this._port = 0;
-    this._devices && this._devices.clear();
-    this._devices = null;
+  },
+
+  // nsIObserver
+  observe: function(aSubject, aTopic, aData) {
+    DEBUG && log("TCPPresentationServer - observe: " + aTopic);
+    switch (aTopic) {
+      case "network-active-changed": {
+        if (!aSubject) {
+          DEBUG && log("No active network");
+          return;
+        }
+
+        /**
+         * Restart service only when original status is online because other
+         * cases will be handled by "network:offline-status-changed".
+         */
+        if (!Services.io.offline) {
+          this._restartService();
+        }
+        break;
+      }
+      case "network:offline-status-changed": {
+        if (aData == "offline") {
+          DEBUG && log("network offline");
+          return;
+        }
+        this._restartService();
+        break;
+      }
+    }
+  },
+
+  _restartService: function() {
+    DEBUG && log("TCPPresentationServer - restart service");
+
+    // restart server socket
+    if (this._isServiceInit()) {
+      let port = this._port;
+      this.close();
+
+      try {
+        this.startService();
+        if (this._listener && this._port !== port) {
+           this._listener.onPortChange(this._port);
+        }
+      } catch (e) {
+        DEBUG && log("TCPPresentationServer - restart service fail: " + e);
+      }
+    }
   },
 
   classID: Components.ID("{f4079b8b-ede5-4b90-a112-5b415a931deb}"),
   QueryInterface : XPCOMUtils.generateQI([Ci.nsIServerSocketListener,
-                                          Ci.nsITCPPresentationServer]),
+                                          Ci.nsITCPPresentationServer,
+                                          Ci.nsIObserver]),
 };
 
 function ChannelDescription(aInit) {
@@ -261,8 +255,8 @@ function ChannelDescription(aInit) {
       this._tcpAddresses = Cc["@mozilla.org/array;1"]
                            .createInstance(Ci.nsIMutableArray);
       for (let address of aInit.tcpAddress) {
-        let wrapper = Cc["@mozilla.org/supports-string;1"]
-                      .createInstance(Ci.nsISupportsString);
+        let wrapper = Cc["@mozilla.org/supports-cstring;1"]
+                      .createInstance(Ci.nsISupportsCString);
         wrapper.data = address;
         this._tcpAddresses.appendElement(wrapper, false);
       }
@@ -310,7 +304,7 @@ function discriptionAsJson(aDescription) {
       let addresses = aDescription.tcpAddress.QueryInterface(Ci.nsIArray);
       json.tcpAddress = [];
       for (let idx = 0; idx < addresses.length; idx++) {
-        let address = addresses.queryElementAt(idx, Ci.nsISupportsString);
+        let address = addresses.queryElementAt(idx, Ci.nsISupportsCString);
         json.tcpAddress.push(address.data);
       }
       json.tcpPort = aDescription.tcpPort;
@@ -324,13 +318,13 @@ function discriptionAsJson(aDescription) {
 
 function TCPControlChannel(presentationServer,
                            transport,
-                           device,
+                           deviceInfo,
                            presentationId,
                            direction,
                            url) {
   DEBUG && log("create TCPControlChannel: " + presentationId + " with role: "
                + direction);
-  this._device = device;
+  this._deviceInfo = deviceInfo;
   this._presentationId = presentationId;
   this._direction = direction;
   this._transport = transport;
@@ -423,6 +417,14 @@ TCPControlChannel.prototype = {
     this._sendMessage("answer", msg);
   },
 
+  sendIceCandidate: function(aCandidate) {
+    let msg = {
+      type: "requestSession:IceCandidate",
+      presentationId: this.presentationId,
+      iceCandidate: aCandidate,
+    };
+    this._sendMessage("iceCandidate", msg);
+  },
   // may throw an exception
   _send: function(aMsg) {
     DEBUG && log("TCPControlChannel - Send: " + JSON.stringify(aMsg, null, 2));
@@ -487,7 +489,7 @@ TCPControlChannel.prototype = {
   onStopRequest: function(aRequest, aContext, aStatus) {
     DEBUG && log("TCPControlChannel - onStopRequest: " + aStatus
                  + " with role: " + this._direction);
-    this.close();
+    this.close(Cr.NS_OK);
     this._notifyClosed(aStatus);
   },
 
@@ -524,13 +526,13 @@ TCPControlChannel.prototype = {
   // Handle command from remote side
   _handleMessage: function(aMsg) {
     DEBUG && log("TCPControlChannel - handleMessage from "
-                 + JSON.stringify(this._device) + ": " + JSON.stringify(aMsg));
+                 + JSON.stringify(this._deviceInfo) + ": " + JSON.stringify(aMsg));
     switch (aMsg.type) {
       case "requestSession:Init": {
-        this._device.id = aMsg.id;
+        this._deviceInfo.id = aMsg.id;
         this._url = aMsg.url;
         this._presentationId = aMsg.presentationId;
-        this._presentationServer.onSessionRequest(aMsg.id,
+        this._presentationServer.onSessionRequest(this._deviceInfo,
                                                   aMsg.url,
                                                   aMsg.presentationId,
                                                   this);
@@ -538,11 +540,19 @@ TCPControlChannel.prototype = {
         break;
       }
       case "requestSession:Offer": {
-        this._listener.onOffer(new ChannelDescription(aMsg.offer));
+        this._onOffer(aMsg.offer);
         break;
       }
       case "requestSession:Answer": {
-        this._listener.onAnswer(new ChannelDescription(aMsg.answer));
+        this._onAnswer(aMsg.answer);
+        break;
+      }
+      case "requestSession:IceCandidate": {
+        this._listener.onIceCandidate(aMsg.iceCandidate);
+        break;
+      }
+      case "requestSession:CloseReason": {
+        this._pendingCloseReason = aMsg.reason;
         break;
       }
     }
@@ -570,7 +580,7 @@ TCPControlChannel.prototype = {
       let offer = this._pendingOffer;
       DEBUG && log("TCPControlChannel - notify pending offer: "
                    + JSON.stringify(offer));
-      this._listener._onOffer(new ChannelDescription(offer));
+      this._listener.onOffer(new ChannelDescription(offer));
       this._pendingOffer = null;
     }
 
@@ -578,7 +588,7 @@ TCPControlChannel.prototype = {
       let answer = this._pendingAnswer;
       DEBUG && log("TCPControlChannel - notify pending answer: "
                    + JSON.stringify(answer));
-      this._listener._onAnswer(new ChannelDescription(answer));
+      this._listener.onAnswer(new ChannelDescription(answer));
       this._pendingAnswer = null;
     }
 
@@ -598,7 +608,7 @@ TCPControlChannel.prototype = {
       return;
     }
     if (!this._listener) {
-      this._pendingOffer = offer;
+      this._pendingOffer = aOffer;
       return;
     }
     DEBUG && log("TCPControlChannel - notify offer: "
@@ -622,7 +632,7 @@ TCPControlChannel.prototype = {
   _notifyOpened: function() {
     this._connected = true;
     this._pendingClose = false;
-    this._pendingCloseReason = null;
+    this._pendingCloseReason = Cr.NS_OK;
 
     if (!this._listener) {
       this._pendingOpen = true;
@@ -640,10 +650,15 @@ TCPControlChannel.prototype = {
     this._pendingOffer = null;
     this._pendingAnswer = null;
 
+    // Remote endpoint closes the control channel with abnormal reason.
+    if (aReason == Cr.NS_OK && this._pendingCloseReason != Cr.NS_OK) {
+      aReason = this._pendingCloseReason;
+    }
+
     if (!this._listener) {
-     this._pendingClose = true;
-     this._pendingCloseReason = aReason;
-     return;
+      this._pendingClose = true;
+      this._pendingCloseReason = aReason;
+      return;
     }
 
     DEBUG && log("TCPControlChannel - notify closed with role: "
@@ -651,9 +666,21 @@ TCPControlChannel.prototype = {
     this._listener.notifyClosed(aReason);
   },
 
-  close: function() {
-    DEBUG && log("TCPControlChannel - close");
+  close: function(aReason) {
+    DEBUG && log("TCPControlChannel - close with reason: " + aReason);
+
     if (this._connected) {
+      // default reason is NS_OK
+      if (typeof aReason !== "undefined" && aReason !== Cr.NS_OK) {
+        let msg = {
+          type: "requestSession:CloseReason",
+          presentationId: this.presentationId,
+          reason: aReason,
+        };
+        this._sendMessage("close", msg);
+        this._pendingCloseReason = aReason;
+      }
+
       this._transport.setEventSink(null, null);
       this._pump = null;
 
@@ -670,43 +697,4 @@ TCPControlChannel.prototype = {
                                          Ci.nsIStreamListener]),
 };
 
-function TCPDevice(aPresentationServer, aInfo) {
-  DEBUG && log("create TCPDevice");
-  this.id = aInfo.id;
-  this.name = aInfo.name;
-  this.type = aInfo.type
-  this.host = aInfo.host;
-  this.port = aInfo.port;
-
-  this._presentationServer = aPresentationServer;
-  this._listener = null;
-}
-
-TCPDevice.prototype = {
-  establishControlChannel: function(aUrl, aPresentationId) {
-    DEBUG && log("TCPDevice - establishControlChannel: " + aUrl + ", "
-                 + aPresentationId);
-    return this._presentationServer
-               .requestSession(this._getDeviceInfo(), aUrl, aPresentationId);
-  },
-  get listener() {
-    return this._listener;
-  },
-  set listener(aListener) {
-    DEBUG && log("TCPDevice - set listener");
-    this._listener = aListener;
-  },
-
-  _getDeviceInfo: function() {
-    return new TCPDeviceInfo(this.host,
-                             this.port,
-                             this.id,
-                             this.name,
-                             this.type);
-  },
-
-  classID: Components.ID("{d6492549-a4f2-4a0c-9a93-00f0e9918b0a}"),
-  QueryInterface: XPCOMUtils.generateQI([Ci.nsIPresentationDevice]),
-};
-
 this.NSGetFactory = XPCOMUtils.generateNSGetFactory([TCPPresentationServer]);
